#include <elf.h>
#include <l4e.h>

/* Checks that elfFile points to a valid elf file.
 * Returns 0 if the elf file is valid, < 0 if invalid.
 */

int
elf_checkFile (void *elfFile)
{
	struct Elf32_Header *fileHdr = (struct Elf32_Header *) elfFile;
	if (fileHdr->e_ident[EI_MAG0] != ELFMAG0
	    || fileHdr->e_ident[EI_MAG1] != ELFMAG1
	    || fileHdr->e_ident[EI_MAG2] != ELFMAG2
	    || fileHdr->e_ident[EI_MAG3] != ELFMAG3)
		return -1; /* not an elf file */
	if (fileHdr->e_ident[EI_CLASS] != ELFCLASS32)
		return -2; /* not 32-bit file */
	if (fileHdr->e_ident[EI_DATA] != 2)
		return -3; /* not big-endian file */
	if (fileHdr->e_ident[EI_VERSION] != 1)
		return -4; /* wrong version of elf */
	if (fileHdr->e_machine != 8)
		return -5; /* wrong architecture (not MIPS) */
	if (fileHdr->e_type != 2)
		return -6; /* not an executable program */
	if (fileHdr->e_phentsize != sizeof(struct Elf32_Phdr))
		return -7; /* unexpected size of program segment header */
	if (fileHdr->e_phnum == 0)
		return -8; /* no program segments */
	if ((fileHdr->e_flags & 0x7e) != 0)
		return -9; /* wrong flags (did you forgot to compile 
			      with -mno-abicalls?)*/

	return 0; /* elf file looks OK */
}

struct Elf32_Phdr *
getProgramSegmentTable (void *elfFile)
/* Returns a pointer to the program segment table, which is an array of
 * ELF64_Phdr_t structs.  The size of the array can be found by calling
 * getNumProgramSegments.
 */
{
	struct Elf32_Header *fileHdr = (struct Elf32_Header *) elfFile;
	return (struct Elf32_Phdr *)(fileHdr->e_phoff + (long)elfFile);
}

unsigned
getNumProgramSegments (void *elfFile)
/* Returns the number of program segments in this elf file.
 */
{
	struct Elf32_Header *fileHdr = (struct Elf32_Header *) elfFile;
	return fileHdr->e_phnum;
}


/* Debugging functions */

/* prints out some details of one elf file */
void
elf_showDetails (void *elfFile, char *name)
{
	struct Elf32_Phdr *segments;
	unsigned numSegments;
	int i;

	/* get a pointer to the table of program segments */
	segments = getProgramSegmentTable (elfFile);
	numSegments = getNumProgramSegments (elfFile);
	
	l4e_printf ("Found an elf file called \"%s\" located "
		    "at address 0x%lx\n",
		    name, elfFile);

	/* print out info about each segment */
	for (i = 0; i < numSegments; i++) {
		if (segments[i].p_type != 1) {
			l4e_printf ("segment %d is not loadable, "
				    "skipping\n", i);
		} else {
			l4e_printf ("segment %d:"
				    " located at file offset 0x%lx, "
				    "size in file is 0x%lx\n"
				    "           load to vaddr 0x%lx, "
				    "expand to size 0x%lx\n", i,
				    segments[i].p_offset, segments[i].p_filesz,
				    segments[i].p_vaddr, segments[i].p_memsz); 
		}
	}
}
