/****************************************************************************
 *
 *      $Id: $
 *
 *      Description: L4 Environment library
 *
 *      Author:      Ben Leslie
 *
 ****************************************************************************/

#include <types.h>
#include <stddef.h>
#include <l4.h>
#include <l4/mips64/kdebug.h>
#include <l4e.h>
#include <stdarg.h>
#include <sprintf.h>

#define TASKBITS 17
#define THREADBITS 10
#define DEBUG 0

L4_ThreadId_t root_thread;

static L4_Fpage_t kip_area, utcb_area;
static void * kip;
static L4_Word_t utcb_base;
static L4_ThreadId_t me;
static L4_Word_t next_no;

/* Initialise the L4 Environment library */
int
l4e_init(void)
{
	L4_ThreadId_t mylocalid;

	/* Find out my ID */
	root_thread = me = L4_Myself ();
	
	/* Find out the kernel information page */
	kip = L4_GetKernelInterface ();
	
	/* and get the correct area */
	kip_area = L4_FpageLog2 ((L4_Word_t) kip, L4_KipAreaSizeLog2 (kip));

#if DEBUG
	l4e_printf("KIP at: ");
	l4e_print_fpage(kip_area);
	l4e_printf("\n");
#endif
	
	/* get my localid to find the utcb base */
	mylocalid  = L4_MyLocalId ();
	utcb_base = *(L4_Word_t *) &mylocalid;
	utcb_base &= ~(L4_UtcbAreaSize (kip) - 1);
	    
	/* Make room for at least 1024 threads in the UTCB area */
	utcb_area = L4_Fpage (utcb_base, L4_UtcbSize (kip) * 1024);
	
#if DEBUG
	l4e_printf("UTCB at: ");
	l4e_print_fpage(utcb_area);
	l4e_printf("\n");
#endif

	/* get a base thread id */
	next_no = L4_ThreadNo (L4_MyGlobalId ()) + 1;

	return 0;
}

/*
 *
 * L4 Debug output functions
 *
 */

/* Output a character to the debug console */
static int
l4e_putc(int c)
{
	L4_KDB_PrintChar( c );
	if( c == '\n' )
		L4_KDB_PrintChar( '\r' );
	return 0;
}

/* Output a string to the debug console */
void
l4e_print_string(char *s)
{
    while ( *s )
	    l4e_putc(*s++);
}

/* Fixed size buffer for debug printf. This is not thread 
   safe (or particularly effecient */

#define PRINTF_SZ 512
static char buf[PRINTF_SZ];

int
l4e_printf(const char *fmt, ...)
{
	va_list         ap;
	int len;
	va_start (ap, fmt);
	len = vsnprintf (buf, PRINTF_SZ, fmt, ap);
	va_end (ap);

	l4e_print_string(buf);

	return 0;
}

/* pretty print an L4 error code */
void
l4e_print_error(L4_Word_t ec)
{
	L4_Word_t e, p;

	/* mmm, seedy */
	e = (ec >> 1) & 7;
	p = ec & 1;

	l4e_printf("IPC %s error, code is 0x%lx (see p60)\n",
		   p ? "receive" : "send", e);
	if (e == 4) {
		l4e_printf("  Message overflow error\n");
	}
	

}

/* pretty print and L4 fpage */
void
l4e_print_fpage(L4_Fpage_t fpage)
{
	l4e_printf("fpage(%lx - %lx)", L4_Address(fpage), 
	      L4_Address(fpage) + L4_Size(fpage));
}



/* Thread and task ID handling function */

/* A simple kernel thread id allocator */
L4_ThreadId_t
l4e_get_new_tid (void)
{
	return L4_GlobalId (next_no++, 1);
}

/* Return tid for a given task */
L4_ThreadId_t 
l4e_task_2_tid(int taskid)
{
	return L4_GlobalId(taskid << THREADBITS, 1);
 }

/* Return the task for a given tid */
int
l4e_tid_2_task(L4_ThreadId_t tid)
{
	return L4_ThreadNo(tid) >> THREADBITS;
}


/* Create a new thread */
static L4_ThreadId_t
create_thread (L4_ThreadId_t tid)
{
    L4_Word_t utcb_location;
    int res;

    me = L4_Myself ();

    utcb_location = utcb_base + L4_UtcbSize (kip) * L4_ThreadNo(tid);

    /* Create active thread */
    res = L4_ThreadControl (tid, me, me, me, 
			    (void *) utcb_location);
    if (res != 1)
	    l4e_printf ("ERROR: ThreadControl returned %d %p\n", 
			res, utcb_location);
    return tid;
}

/* Start a thread executing */
static void
start_thread( L4_ThreadId_t tid, L4_Word_t ip, L4_Word_t sp )
{
	L4_Msg_t msg;
	L4_MsgTag_t tag;

	L4_MsgClear(&msg);
	L4_MsgAppendWord(&msg, ip);
	L4_MsgAppendWord(&msg, sp);

	/*
	  The start thread message must come from the thread's pager.
	   Here we make the gross assumption that the thread's pager is
	   "me", which is the initial thread
	*/
	L4_Set_Propagation(&msg.tag);
	L4_Set_VirtualSender(me);

	L4_MsgLoad(&msg);

	tag = L4_Send(tid);

	if (L4_IpcFailed(tag)) {
		l4e_printf("emu: starting thread failed: %lx\n", tag.raw);
		l4e_print_error(L4_ErrorCode());
	}
}

/* Create and start a thread */
int
l4e_start_thread(L4_ThreadId_t tid, void *entrypoint, void *stack)
{
	/* This bit creates the thread, but it won't execute any
	   code yet 
	*/
	create_thread(tid);

	/* Send an ipc to thread thread to start it up */
	start_thread(tid, (L4_Word_t) entrypoint, (L4_Word_t) stack);

	return 0;
}

/* Create and start a new task */
int
l4e_task_new(L4_ThreadId_t task, L4_ThreadId_t pager, 
	     void *entrypoint, void *stack)
{
	L4_Word_t control;
	int res;

	/* Create an inactive thread */
	res = L4_ThreadControl (task, task, me, L4_nilthread, (void *) -1);
	if (res != 1)
		return -1;

	/* Setup space */
	res = L4_SpaceControl (task, 0, kip_area, utcb_area,
			       L4_nilthread, &control);
	if (res != 1)
		return -2;

	/* Activate thread */
	res = L4_ThreadControl (task, task, me, pager, (void *) utcb_base);
	if (res != 1)
		return -3;

	start_thread(task, (L4_Word_t) entrypoint, (L4_Word_t) stack);

	return 0;
}

/* Do an explicit request to sigma0 */

#define FP_SSHIFT 4    /* shift amt for 's' part */
#define FP_RD (1<<2)   /* Read, write and ex bits... */
#define FP_WR (1<<1)  
#define FP_EX (1<<0)
#define FPAGE_MASK (~(2<<10)-1)  /* everything but lowest 10 bits */
#define FP_REQUEST_LABEL ((-6UL) << 4)     /* that's what the API says! */

int
l4e_request_page( void *physaddr, int cached )
{
	L4_ThreadId_t tid, from;
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_Word_t raddr, rfpage, rattrib;
	L4_MapItem_t map;
	physaddr = (void*) ((((L4_Word_t) physaddr) >> PAGEBITS) << PAGEBITS);

	/* find our pager's ID */
	tid = L4_Pager();

	/* setup the request */
	raddr = ((L4_Word_t) physaddr) & FPAGE_MASK;
	rfpage = raddr | (PAGEBITS << FP_SSHIFT) | FP_RD | FP_WR | FP_EX;
	rattrib = 0;  /* arch. default attributes */

	if (cached == 0) {
		l4e_printf("Requesting uncached mappings: %x\n",
			   raddr);
		rattrib = 1;
	}

	/* accept fpages */
	L4_Accept(L4_MapGrantItems(L4_CompleteAddressSpace));

	/* send it to our pager */
	L4_MsgClear( &msg );
	L4_MsgAppendWord(&msg, rfpage );
	L4_MsgAppendWord(&msg, rattrib );
	L4_Set_Label( &msg.tag, FP_REQUEST_LABEL );
	L4_MsgLoad( &msg );

	/* make the call */
	tag = L4_Ipc(tid, tid, L4_Timeouts(L4_Never, L4_Never), &from);

	/* check for an error */
	if (L4_IpcFailed(tag)) {
		//L4_Word_t ec = L4_ErrorCode();
		l4e_printf("l4_request_page: IPC error\n");
		//l4e_print_error(ec);
		return 0x37;
	}

	/* FIXME: check no. typed/untyped words? */
	/* decipher the results */
	L4_MsgStore(tag, &msg);
	L4_MsgGetMapItem( &msg, 0, &map );

	/* rejected mapping? */
	if( map.X.snd_fpage.raw == L4_Nilpage.raw ) {
		l4e_printf("l4_request_page: no fpage received %p\n", physaddr);
		return 1;
	}

	return 0;
}

/* Flush the L4 caches */
void
l4e_cache_flush(void)
{
	int r;
	L4_Word_t attrs[4];

	r = L4_MemoryControl( -2UL, attrs );
}


/* Find the largest available chunk of physical RAM */
void
l4e_find_memory(L4_Word_t *low, L4_Word_t *high)
{
	int i;
	L4_KernelInterfacePage_t *kip;
	L4_MemoryDesc_t *memdesc;
	L4_Word_t mem_begin = 0, mem_end = 0;
	char *type;

	kip = L4_GetKernelInterface();

	for (i = 0; i < kip->MemoryInfo.n; i++) {
		memdesc = L4_MemoryDesc(kip, i);

		*low = L4_MemoryDescLow(memdesc);
		*high = L4_MemoryDescHigh(memdesc);

		/* L4 uses inclusive addresses, we prefer
		 * to label the next byte after.
		 */
	        (*high)++;

		switch (memdesc->x.type) {
			case 0:  type = "undefined";	break;
			case 1:  type = "conventional";	break;
			case 2:  type = "reserved";	break;
			case 3:  type = "dedicated";	break;
			default: type = "ERROR";	break;
		}


		/* see if it looks like main memory (dodgy, FIXME) */
		if (memdesc->x.type == 1 && memdesc->x.v == 0
		    && (*high - *low) > (1 * 1024 * 1024)) {
			mem_begin = *low;
			mem_end = *high;
		}

		/* make sure we don't use any reserved segments
		 * FIXME: do the reserved sections always come after 
		 * conventional? if not, this will break */
		if (memdesc->x.type == 2 && memdesc->x.v == 0) {
			/* calculate remaining sizes on each side */
			L4_Word_t size1 = 0;
			L4_Word_t size2 = 0;
			if ((*low >= mem_begin) && (*low < mem_end))
				size1 = *low - mem_begin;
			if ((*high > mem_begin) && (*high <= mem_end))
				size2 = mem_end - *high;
			if (size1 > size2)
				mem_end = *low;
			else if (size2 > 0)
				mem_begin = *high;
		}
	}
	*high = mem_end;
	*low = mem_begin;
	
}


