#include <l4.h>
#include <l4e.h>
#include <dev.h>
#include <stddef.h>
#include <serial.h>
#include <io.h>
#include <assert.h>
#include <mutex.h>

#include "z85230.h"

static inline void wbflush(void)
{
    asm ("sync;\n\t");
}

#define CBAUD   0000017
#define B0      0
#define B50     0000001
#define B75     0000002
#define B110    0000003
#define B134    0000004
#define B150    0000005
#define B200    0000006
#define B300    0000007
#define B600    0000010
#define B1200   0000011
#define B1800   0000012
#define B2400   0000013
#define B4800   0000014
#define B9600   0000015
#define B19200  0000016
#define B38400  0000017

#define CPUCLK  100000000	 /* CPU clock rate 100 MHz */
#define ZSCCCLK	14745600	 /* ZSCC clock rate 14.7456 MHz */

/* experimental evidence dictates that we need to wait a few ZSCC clock
 * cycles between register reads and writes. the following works well.
 */
#define DELAY	{ volatile int i; for (i = 0; i < 3; i++); }

#define FIFO_SIZE	1024
#define FIFO_SIZE_MASK	(FIFO_SIZE-1)

struct serial {
	bus_space_t cmd_port;
	bus_space_t data_port;
	unsigned char output_fifo[FIFO_SIZE];
	void (*handler) (char c);
	volatile int wnd_lo, wnd_hi;
	volatile int async;
};


static const int zsccbrtc[] = {
    -1,
    BRTC (ZSCCCLK, 50),
    BRTC (ZSCCCLK, 75),
    BRTC (ZSCCCLK, 110),
    BRTC (ZSCCCLK, 134),
    BRTC (ZSCCCLK, 150),
    BRTC (ZSCCCLK, 200),
    BRTC (ZSCCCLK, 300),
    BRTC (ZSCCCLK, 600),
    BRTC (ZSCCCLK, 1200),
    BRTC (ZSCCCLK, 1800),
    BRTC (ZSCCCLK, 2400),
    BRTC (ZSCCCLK, 4800),
    BRTC (ZSCCCLK, 9600),
    BRTC (ZSCCCLK, 19200),
    BRTC (ZSCCCLK, 38400)
};

static void
zsccputreg (struct serial *serial, unsigned char reg, unsigned char val)
{
	if (reg != zWR0) {
		DELAY;
		io_write_8(serial->cmd_port, 0, zWR0_REG | reg);
		wbflush();
	}
	DELAY;
	io_write_8(serial->cmd_port, 0, val);
	wbflush();
}

static unsigned char
zsccgetreg (struct serial *serial, unsigned char reg)
{
	if (reg != zRR0) {
		DELAY;
		io_write_8(serial->cmd_port, 0, zWR0_REG | reg);
		wbflush();
	}
	DELAY;
	return io_read_8(serial->cmd_port, 0);
}

static int
zsccprogram (struct serial *serial, int baudrate)
{
	baudrate &= CBAUD;
	if (baudrate == 0)
		return 1;
	
	/*
	 * See the zscc manual for details.
	 */

	zsccputreg(serial, zWR4, zWR4_1STOPBIT | zWR4_CLK);
	zsccputreg(serial, zWR10, zWR10_NRZ);
	
	zsccputreg(serial, zWR14, zWR14_NOP); /* stop BRG */
	zsccputreg(serial, zWR11, zWR11_TRXCOUT| zWR11_TRXCBRG| 
		   zWR11_TCLKBRG | zWR11_RCLKBRG);
	zsccputreg(serial, zWR12,  zsccbrtc [baudrate] & 0xff);
	zsccputreg(serial, zWR13,  (zsccbrtc [baudrate] >> 8) & 0xff);
	
	zsccputreg(serial, zWR14, zWR14_BRGENABLE | zWR14_BRGSRC );
	
	zsccputreg(serial, zWR15, 1);
	zsccputreg(serial, zWR7, zWR7P_TXEMPTYIEN);

	zsccputreg(serial, zWR1, zWR1_TXIEN | zWR1_RXIENALL);
	zsccputreg(serial, zWR9, zWR9_MIE | zWR9_INTACK);

	zsccputreg(serial, zWR2, 0);
	
	zsccputreg(serial, zWR3, zWR3_RXENABLE | zWR3_RX8BITCHAR);
	zsccputreg(serial, zWR5, zWR5_TXENABLE| zWR5_RTS | 
		   zWR5_TX8BITCHAR | zWR5_DTR);
	

	return 0;
}

/* Transmit from FIFO */
static void 
fifo_transmit(struct serial *serial)
{
	while (1) {
		if (serial->wnd_lo == serial->wnd_hi) {
			/* nothing more for now */
			serial->async = 0;
			break;
		}

		if (!(zsccgetreg(serial, zRR0) & zRR0_TXEMPTY)) {
			/* hardware FIFO full - go serial->asynchronous */
			serial->async = 1;
			break;
		}

		io_write_8(serial->data_port, 0, 
			   serial->output_fifo[serial->wnd_lo]);
		wbflush();

		serial->wnd_lo++;
		serial->wnd_lo &= FIFO_SIZE_MASK;
	}
}

/* Queue data into FIFO */
int
serial_send(struct serial *serial, char *data, int len)
{
	int next_hi = serial->wnd_hi;
	int newline = 0;
	int sent = 0;

	while (sent != len) {
		next_hi++;
		next_hi &= FIFO_SIZE_MASK;
		
		if (next_hi == serial->wnd_lo)
			break;
		
		if ((*data == '\n') && !newline) {
			/* dynamically translate LF into CR-LF */
			serial->output_fifo[serial->wnd_hi] = '\r';
			newline = 1;
		} else {
			serial->output_fifo[serial->wnd_hi] = *(data++);
			newline = 0;
			sent++;
		}
		
		serial->wnd_hi = next_hi;
	}
	
	if (!serial->async) {
		fifo_transmit(serial);
	}

	return sent;
}


struct serial *
serial_init(bs_seq spaces)
{
	static struct serial serial; 
	
	serial.cmd_port = spaces.v[0];
	serial.data_port = spaces.v[1];
	serial.async = serial.wnd_hi = serial.wnd_lo = 0;
	serial.handler = NULL;
	zsccprogram(&serial, B38400);
	
	return &serial;
}

int
serial_register_handler(struct serial *serial, void (*handler) (char c))
{
	serial->handler = handler;
	return 0;
}

int
serial_interrupt_handler(struct serial *serial)
{
	int vector = zsccgetreg(serial, zRR2B);
	int ip = (vector >> 1) & 0x3;

	if ((vector >> 4) & 0x1) {
		l4e_printf("Channel A irq %x\n", vector);
		zsccputreg(serial, zWR0, zWR0_RESETIUS);
		return 0;
	}

	if (ip == 0) {
		if (serial->async) {
			fifo_transmit(serial);
		} else {
			zsccputreg(serial, zWR0, zWR0_RESETTXINT);
		}
	}

	if (ip == 1) {
		/* Status interrupt - ignore */
		//int stat = 
			zsccgetreg(serial, zWR0);
		//l4e_printf("Status: %x %x\n", stat, pend);
	}

	if (ip == 2) {
		char c;
		c = io_read_8(serial->data_port, 0);
		if (c == '\r') c = '\n';

		if (serial->handler != NULL) 
			serial->handler(c);
	} 

	if (ip == 3) {
		//int err_vector = zsccgetreg(serial, zRR1);
		//l4e_printf("Error reset %x\n", err_vector);
		zsccputreg(serial, zWR0, zWR0_ERRORRESET);
	}

	/* reset interrupt under service bits */
	zsccputreg(serial, zWR0, zWR0_RESETIUS);
	
	return 0;
}
